package org.hidetake.elshard.demo.controller.tweet.fetch;

import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletResponse;

import org.hidetake.elshard.demo.model.Tweet;
import org.hidetake.elshard.demo.service.TweetParseJobnetService;
import org.hidetake.elshard.demo.service.TweetService;
import org.hidetake.elshard.task.Configuration;
import org.hidetake.elshard.task.Context;
import org.hidetake.elshard.task.s3.TaskChainController;
import org.slim3.controller.validator.Validators;
import org.slim3.util.IntegerUtil;

import twitter4j.Paging;
import twitter4j.TwitterException;

/**
 * タスクチェーンによりツイートを取得する．
 * @author hidetake.org
 */
public class ChainController extends TaskChainController
{

	private static final Logger logger = Logger.getLogger(ChainController.class.getName());
	private static final String PARAM_USERNAME = "userName";

	private final TweetService tweetService = new TweetService();
	private final TweetParseJobnetService tweetParseJobnetService = new TweetParseJobnetService();

	@Override
	protected void configure(Configuration configuration)
	{
	}

	@Override
	public List<String> input(Context context) throws Exception
	{
		if(!validate()) {
			return applicationError(errors.toString());
		}
		
		// 取得中か？
		String userName = param(PARAM_USERNAME);
		Date since = tweetParseJobnetService.lock(userName);
		if(since != null) {
			return applicationError("Fetch in progress since "+since.toString());
		}
		
		// 1ページ目
		try {
			Paging paging = new Paging(1, 100);
			List<Tweet> tweets = tweetService.fetchAndSaveTimeline(userName, paging);
			if(tweets.isEmpty()) {
				return applicationError("No tweet found.");
			}
		}
		catch(TwitterException e) {
			return applicationError(e.getLocalizedMessage());
		}
		
		// 2ページ目のタスクを追加する．
		context.addParameter(PARAM_USERNAME, userName);
		return Arrays.asList("2");
	}

	@Override
	public List<String> run(List<String> input, Context context) throws Exception
	{
		String userName = param(PARAM_USERNAME);
		int pageIndex = IntegerUtil.toPrimitiveInt(input.get(0));
		Paging paging = new Paging(pageIndex, 100);

		try {
			List<Tweet> tweets = tweetService.fetchAndSaveTimeline(userName, paging);
			if(tweets.isEmpty()) {
				logger.info("ページ "+pageIndex+" をリクエストしたけど，これ以上ページがないとです．");
				tweetParseJobnetService.startMapperJob(userName);
				return null;
			}
			
			logger.info("ページ "+pageIndex+" のツイート "+tweets.size()+" 件を取得しました．");
			context.addParameter(PARAM_USERNAME, userName);
			return Arrays.asList(Integer.toString(pageIndex + 1));
		}
		catch(TwitterException e) {
			// 通信エラーはリトライする．
			logger.warning(e.getLocalizedMessage());
			context.addParameter(PARAM_USERNAME, userName);
			return Arrays.asList(Integer.toString(pageIndex));
		}
	}

	protected List<String> applicationError(String message) throws Exception
	{
		response.setStatus(HttpServletResponse.SC_BAD_REQUEST);
		response.setContentType("text/plain");
		response.setCharacterEncoding("UTF-8");
		response.getWriter().append(message);
		return null;
	}

	private boolean validate()
	{
		Validators v = new Validators(request);
		v.add(PARAM_USERNAME, v.required(), v.maxlength(50));
		return v.validate();
	}

}
