package org.hidetake.elshard.task;

import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.servlet.http.HttpServletRequest;

import org.hidetake.elshard.task.impl.ConfigurationImpl;
import org.hidetake.elshard.task.impl.ContextImpl;
import org.hidetake.elshard.util.RequestUtil;

import com.google.appengine.api.taskqueue.TaskHandle;
import com.google.appengine.api.taskqueue.TaskOptions;

/**
 * The request handler to execute task chain.
 * 
 * @author hidetake.org
 */
public class TaskChainRequestHandler
{

	private static final String PARAM_TASKCHAIN_RESULT = "X-TaskChain-Result";
	private static final String HEADER_TASKCHAIN_ID = "X-TaskChain-ID";
	private static final String HEADER_TASKCHAIN_SEQUENCE = "X-TaskChain-Sequence";

	private final Input input;
	private final Task task;
	private final ConfigurationImpl configuration;

	/**
	 * Constructor with default configuration.
	 * @param input {@link Input} object
	 * @param task {@link Task} object
	 */
	public TaskChainRequestHandler(Input input, Task task)
	{
		this.input = input;
		this.task = task;
		this.configuration = new ConfigurationImpl();
	}

	/**
	 * Constructor.
	 * @param input {@link Input} object
	 * @param task {@link Task} object
	 * @param configuration configuration of the task chain
	 */
	public TaskChainRequestHandler(Input input, Task task, ConfigurationImpl configuration)
	{
		this.input = input;
		this.task = task;
		this.configuration = configuration;
	}

	/**
	 * Get the task chain configuration.
	 * @return
	 */
	public Configuration getConfiguration()
	{
		return configuration;
	}

	/**
	 * Handle a HTTP request.
	 * Any {@link Throwable} will be caught to prevent from infinite loop.
	 * @param request
	 */
	public void handle(HttpServletRequest request)
	{
		try {
			if(request.getParameterValues(PARAM_TASKCHAIN_RESULT) == null) {
				handleInput(request);
			}
			else {
				handleTask(request);
			}
		}
		catch(Throwable e) {
			Logger logger = Logger.getLogger(task.getClass().getName());
			logger.log(Level.SEVERE, e.getLocalizedMessage(), e);
		}
	}

	/**
	 * Handler for {@link Input} task.
	 * @param request
	 * @throws Exception
	 */
	private void handleInput(HttpServletRequest request) throws Exception
	{
		ContextImpl context = new ContextImpl(
				request,
				configuration,
				request.getRequestURI(),
				Long.toString(System.currentTimeMillis()),
				1);
		
		List<String> result = input.input(context);
		if(result != null && result.size() > 0) {
			List<List<String>> pieces = configuration.getSplitter().split(result);
			for(List<String> piece : pieces) {
				enqueue(piece, context);
			}
		}
	}

	/**
	 * Handler for {@link Task} task.
	 * @param request
	 * @throws Exception
	 */
	private void handleTask(HttpServletRequest request) throws Exception
	{
		if(request.getHeader("X-AppEngine-TaskName") == null) {
			throw new IllegalStateException("Not run as a task");
		}
		
		ContextImpl context = new ContextImpl(
				request,
				configuration,
				request.getRequestURI(),
				RequestUtil.getHeaderAsString(request, HEADER_TASKCHAIN_ID),
				RequestUtil.getHeaderAsInt(request, HEADER_TASKCHAIN_SEQUENCE));
		
		List<String> previousResult = Arrays.asList(request.getParameterValues(PARAM_TASKCHAIN_RESULT));
		List<String> result = task.run(previousResult, context);
		if(result != null && result.size() > 0) {
			List<List<String>> pieces = configuration.getSplitter().split(result);
			for(List<String> piece : pieces) {
				enqueue(piece, context);
			}
		}
	}

	/**
	 * Enqueue a task.
	 * @param context task context
	 * @return the task handle
	 */
	private TaskHandle enqueue(List<String> result, ContextImpl context)
	{
		StringBuilder taskName = new StringBuilder();
		taskName.append(task.getClass().getSimpleName());
		taskName.append('-');
		taskName.append(context.getTaskChainId());
		taskName.append('-');
		taskName.append(context.getSequenceId());
		taskName.append('-');
		taskName.append(context.incrementTaskCount());
		
		TaskOptions task = TaskOptions.Builder
				.withMethod(TaskOptions.Method.POST)
				.url(context.getTaskUrl())
				.header(HEADER_TASKCHAIN_ID, context.getTaskChainId())
				.header(HEADER_TASKCHAIN_SEQUENCE, Integer.toString(context.getSequenceId()+1))
				.taskName(taskName.toString());
		for(String element : result) {
			task.param(PARAM_TASKCHAIN_RESULT, element);
		}
		for(Map.Entry<String, String> e : context.getParams().entrySet()) {
			task.param(e.getKey(), e.getValue());
		}
		return configuration.getQueue().add(task);
	}

}
