package org.hidetake.elshard.task.s3;

import java.io.Serializable;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.hidetake.elshard.task.Context;
import org.hidetake.elshard.task.Input;
import org.hidetake.elshard.task.Task;
import org.hidetake.elshard.task.TaskChainRequestHandler;
import org.hidetake.elshard.task.splitter.SizeSplitter;
import org.slim3.controller.Controller;
import org.slim3.controller.Navigation;
import org.slim3.datastore.Datastore;
import org.slim3.datastore.ModelQuery;
import org.slim3.datastore.S3QueryResultList;
import org.slim3.memcache.Memcache;

/**
 * Query and Processor task.
 * 
 * @author hidetake.org
 * @param <M> model class (must be {@link Serializable})
 */
public abstract class QueryProcessorController<M extends Serializable> extends Controller implements Input, Task
{

	private static final String PARAM_MODEL = "m";
	private static final String PARAM_SORTS = "s";
	private static final String PARAM_FILTERS = "f";
	private static final String PARAM_CURSOR_CURRENT = "c";
	private static final String PARAM_CURSOR_NEXT = "n";
	
	private final QueryProcessorConfigurationImpl configuration = new QueryProcessorConfigurationImpl();
	private final Map<String, Task> handlerMap = new HashMap<String, Task>();
	
	{
		handlerMap.put(FirstQuery.class.getSimpleName(), new FirstQuery());
		handlerMap.put(Query.class.getSimpleName(), new Query());
		handlerMap.put(FirstProcessor.class.getSimpleName(), new FirstProcessor());
		handlerMap.put(Processor.class.getSimpleName(), new Processor());
	}
	
	/**
	 * Configure the task chain.
	 * @return the configuration
	 */
	protected abstract void configure(QueryProcessorConfiguration configuration);

	/**
	 * 
	 * @param context task chain context
	 * @return query
	 */
	public abstract ModelQuery<M> query(Context context);

	/**
	 * Process data set.
	 * @param input entities
	 * @param context
	 */
	public abstract void process(List<M> input, Context context);

	/*
	 * (non-Javadoc)
	 * @see org.slim3.controller.Controller#run()
	 */
	@Override
	protected final Navigation run() throws Exception
	{
		TaskChainRequestHandler handler = new TaskChainRequestHandler(this, this, configuration);
		configuration.setSplitter(new SizeSplitter(1));
		configure(configuration);
		handler.handle(request);
		return null;
	}

	/*
	 * (non-Javadoc)
	 * @see org.hidetake.elshard.task.Input#input(org.hidetake.elshard.task.Context)
	 */
	@Override
	public List<String> input(Context context) throws Exception
	{
		return Arrays.asList(FirstQuery.class.getSimpleName());
	}

	/*
	 * (non-Javadoc)
	 * @see org.hidetake.elshard.task.Task#run(java.util.List, org.hidetake.elshard.task.Context)
	 */
	@Override
	public List<String> run(List<String> input, Context context) throws Exception
	{
		if(input.size() != 1) {
			throw new IllegalArgumentException("Parameter must be only one");
		}
		Task task = handlerMap.get(input.get(0));
		if(task == null) {
			throw new IllegalArgumentException("Not implemented state: "+input.get(0));
		}
		return task.run(input, context);
	}

	private class FirstQuery implements Task
	{
		@Override
		public List<String> run(List<String> input, Context context) throws Exception
		{
			S3QueryResultList<M> resultList = query(context)
					.limit(configuration.getShardSize())
					.asQueryResultList();
			if(resultList.size() > 0) {
				Memcache.put(context.getTaskChainId()+"first", resultList);
				context.addParameter(PARAM_CURSOR_NEXT, resultList.getEncodedCursor());
				context.addParameter(PARAM_FILTERS, resultList.getEncodedFilters());
				context.addParameter(PARAM_SORTS, resultList.getEncodedSorts());
				context.addParameter(PARAM_MODEL, resultList.get(0).getClass().getName());
				return Arrays.asList(
						Query.class.getSimpleName(),
						FirstProcessor.class.getSimpleName());
			}
			return null;
		}
	}

	private class FirstProcessor implements Task
	{
		@Override
		public List<String> run(List<String> input, Context context) throws Exception
		{
			S3QueryResultList<M> resultList = Memcache.get(context.getTaskChainId()+"first");
			if(resultList == null) {
				resultList = query(context)
						.limit(configuration.getShardSize())
						.asQueryResultList();
			}
			process(resultList, context);
			return null;
		}
	}

	private class Query implements Task
	{
		@Override
		public List<String> run(List<String> input, Context context) throws Exception
		{
			String currentCursor = param(PARAM_CURSOR_NEXT);
			@SuppressWarnings("unchecked")
			Class<M> modelClass = (Class<M>) Class.forName(param(PARAM_MODEL));
			S3QueryResultList<M> resultList = Datastore.query(modelClass)
					.encodedStartCursor(currentCursor)
					.encodedFilters(param(PARAM_FILTERS))
					.encodedSorts(param(PARAM_SORTS))
					.limit(configuration.getShardSize())
					.asQueryResultList();
			if(resultList.size() > 0) {
				Memcache.put(context.getTaskChainId()+currentCursor, resultList);
				context.addParameter(PARAM_CURSOR_CURRENT, currentCursor);
				context.addParameter(PARAM_CURSOR_NEXT, resultList.getEncodedCursor());
				context.addParameter(PARAM_FILTERS, resultList.getEncodedFilters());
				context.addParameter(PARAM_SORTS, resultList.getEncodedSorts());
				context.addParameter(PARAM_MODEL, resultList.get(0).getClass().getName());
				return Arrays.asList(
						Query.class.getSimpleName(),
						Processor.class.getSimpleName());
			}
			return null;
		}
	}

	private class Processor implements Task
	{
		@Override
		public List<String> run(List<String> input, Context context) throws Exception
		{
			String currentCursor = param(PARAM_CURSOR_CURRENT);
			S3QueryResultList<M> resultList = Memcache.get(context.getTaskChainId()+currentCursor);
			if(resultList == null) {
				@SuppressWarnings("unchecked")
				Class<M> modelClass = (Class<M>) Class.forName(param(PARAM_MODEL));
				resultList = Datastore.query(modelClass)
						.encodedStartCursor(currentCursor)
						.encodedFilters(param(PARAM_FILTERS))
						.encodedSorts(param(PARAM_SORTS))
						.limit(configuration.getShardSize())
						.asQueryResultList();
			}
			process(resultList, context);
			return null;
		}
	}

}
