/*
 * Copyright (C) 2010 hidetake.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.hidetake.jsonizer;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

/**
 * オブジェクトをJSON形式に出力するクラス．
 * 
 * オブジェクトは下記のルールで変換される．
 * 
 * <table>
 * <tr><th>Javaオブジェクト</th><th>JSON</th></tr>
 * <tr><td>null</td><td>null</td></tr>
 * <tr><td>{@link String}</td><td>即値</td></tr>
 * <tr><td>char, {@link Character}</td><td>即値</td></tr>
 * <tr><td>boolean, {@link Boolean}</td><td>即値</td></tr>
 * <tr><td>byte, {@link Byte}</td><td>即値</td></tr>
 * <tr><td>short, {@link Short}</td><td>即値</td></tr>
 * <tr><td>int, {@link Integer}</td><td>即値</td></tr>
 * <tr><td>long, {@link Long}</td><td>即値</td></tr>
 * <tr><td>float, {@link Float}</td><td>即値</td></tr>
 * <tr><td>配列</td><td>[e1,e2,...]</td></tr>
 * <tr><td>コレクション</td><td>[e1,e2,...]</td></tr>
 * <tr><td>オブジェクト（クラスに対応する{@link PropertyGetter}がある場合）</td><td>{"key1":"value1","key2":"value2",...}</td></tr>
 * <tr><td>オブジェクト（クラスに対応する{@link PropertyGetter}がない場合）</td><td>{@link Object#toString()}</td></tr>
 * </table>
 * 
 * すべての文字列はエスケープ処理されてからダブルクォートで囲まれる．
 * エスケープ処理の対象を下記に示す．
 * 
 * <ul>
 * <li>バックスラッシュ（\）</li>
 * <li>ダブルクォート（"）</li>
 * <li>改行（\n）</li>
 * </ul>
 */
public final class JsonWriter
{

	private final Appendable appendable;
	private final Map<Class<?>, PropertyGetter<?>> propertyGetterMap;

	/**
	 * {@link PropertyGetter} 無しのJsonWriterを生成する．
	 * @param appendable 出力先
	 */
	public JsonWriter(Appendable appendable)
	{
		this(appendable, Collections.<Class<?>, PropertyGetter<?>>emptyMap());
	}

	/**
	 * JsonWriterを生成する．
	 * @param appendable 出力先
	 * @param propertyGetterMap クラスとそれに対応する {@link PropertyGetter} のマップ
	 */
	public JsonWriter(Appendable appendable, Map<Class<?>, PropertyGetter<?>> propertyGetterMap)
	{
		this.appendable = appendable;
		this.propertyGetterMap = propertyGetterMap;
	}

	/**
	 * オブジェクトを書き出す．
	 * @param object オブジェクト
	 * @throws IOException
	 */
	public final void write(Object object) throws IOException
	{
		writeObject(object);
	}

	/**
	 * オブジェクトを書き出す．
	 * @param object オブジェクト
	 * @throws IOException
	 */
	private final void writeObject(Object object) throws IOException
	{
		if(object == null) {
			appendable.append("null");
		}
		else if(object instanceof String) {
			appendable.append('"');
			appendable.append(JsUtil.escape((String) object));
			appendable.append('"');
		}
		else if(object instanceof Character) {
			appendable.append('"');
			appendable.append(JsUtil.escape((Character) object));
			appendable.append('"');
		}
		else if(object instanceof Boolean ||
				object instanceof Byte ||
				object instanceof Short ||
				object instanceof Integer ||
				object instanceof Long ||
				object instanceof Float ||
				object instanceof Double) {
			appendable.append(object.toString());
		}
		else if(object.getClass().isArray()) {
			writeArray((Object[]) object);
		}
		else if(object instanceof Collection<?>) {
			writeCollection((Collection<?>) object);
		}
		else {
			writeAnother(object);
		}
	}
	
	private <M> void writeAnother(M object) throws IOException
	{
		PropertyGetter<?> exactMatch = propertyGetterMap.get(object.getClass());
		if(exactMatch != null) {
			// exactly matched
			@SuppressWarnings("unchecked")
			PropertyGetter<M> propertyGetter = (PropertyGetter<M>) exactMatch;
			writeHash(object, propertyGetter);
		}
		else {
			for(Entry<Class<?>, PropertyGetter<?>> entry : propertyGetterMap.entrySet()) {
				if(entry.getKey().isInstance(object)) {
					// instance-of test matched
					@SuppressWarnings("unchecked")
					PropertyGetter<M> propertyGetter = (PropertyGetter<M>) entry.getValue();
					writeHash(object, propertyGetter);
					return;
				}
			}
			// no suitable found
			appendable.append('"');
			appendable.append(JsUtil.escape(object.toString()));
			appendable.append('"');
		}
	}
	
	/**
	 * ハッシュを書き出す．
	 * @param object Javaオブジェクト
	 * @param propertyGetter {@link PropertyGetter}
	 * @throws IOException
	 */
	private <M> void writeHash(M object, PropertyGetter<M> propertyGetter) throws IOException
	{
		appendable.append('{');
		List<String> keys = new ArrayList<String>();
		List<Object> values = new ArrayList<Object>();
		propertyGetter.setInternalKeys(keys);
		propertyGetter.setInternalValues(values);
		propertyGetter.get(object);
		if(keys.size() != values.size()) {
			throw new IllegalStateException("key and value don't match at "
					+ propertyGetter.getClass().getName() + "#get()");
		}
		Iterator<String> keyIterator = keys.iterator();
		Iterator<Object> valueIterator = values.iterator();
		if(keyIterator.hasNext()) {
			for(;;) {
				appendable.append('"');
				appendable.append(JsUtil.escape(keyIterator.next()));
				appendable.append("\":");
				writeObject(valueIterator.next());
				if(keyIterator.hasNext()) {
					appendable.append(',');
				}
				else {
					break;
				}
			}
		}
		
		appendable.append('}');
	}

	/**
	 * 配列を書き出す．
	 * 書式は [a,b,...] となる．
	 * @param array Java配列
	 * @throws IOException
	 */
	private void writeArray(Object[] array) throws IOException
	{
		appendable.append('[');
		int length = array.length;
		if(length != 0) {
			for(int i = 0;;) {
				writeObject(array[i]);
				if(++i < length) {
					appendable.append(',');
				}
				else {
					break;
				}
			}
		}
		appendable.append(']');
	}

	/**
	 * コレクションを書き出す．
	 * 書式は [a,b,...] となる．
	 * @param collection Javaコレクション
	 * @throws IOException
	 */
	private void writeCollection(Collection<?> collection) throws IOException
	{
		appendable.append('[');
		Iterator<?> iterator = collection.iterator();
		if(iterator.hasNext()) {
			for(;;) {
				writeObject(iterator.next());
				if(iterator.hasNext()) {
					appendable.append(',');
				}
				else {
					break;
				}
			}
		}
		appendable.append(']');
	}

}
