/*
 * Copyright (C) 2010 hidetake.org
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.hidetake.jsonizer;

import java.util.List;


/**
 * Javaオブジェクトのフィールドを取得してプロパティに格納するクラス．
 * @param <M> 対象クラス
 */
public abstract class PropertyGetter<M>
{

	private static final class EmptyGetter<M> extends PropertyGetter<M>
	{
		@Override
		public void get(M m) {}
	}

	protected final class AsElement
	{
		/**
		 * プロパティのキーを定義する．
		 * @param key キー
		 */
		public final void as(CharSequence key)
		{
			keys.add(key.toString());
		}
		
		private AsElement() {}
	}

	private final AsElement asElement = new AsElement();
	private List<String> keys;
	private List<Object> values;

	/**
	 * プロパティを何も取得しない {@link PropertyGetter} を返す．
	 * @return
	 */
	public static final <M> PropertyGetter<M> emptyGetter()
	{
		return new EmptyGetter<M>();
	}

	/**
	 * オブジェクトからプロパティをgetする．
	 * このメソッドは以下のように実装する．
	 * <code><pre>
	 * public void get(M m) {
	 *   // key value style
	 *   property("element1", m.getElement1());
	 *   property("element2", m.getElement2());
	 *   
	 *   // on the other hand, value as key style
	 *   property(m.getElement1()).as("element1");
	 *   property(m.getElement2()).as("element2");
	 * }
	 * </pre></code>
	 * @param e オブジェクト
	 */
	public abstract void get(M e);

	/**
	 * プロパティを定義する．
	 * @param key キー
	 * @param value 値
	 * @see #property(Object)
	 */
	protected final void property(CharSequence key, Object value)
	{
		keys.add(key.toString());
		values.add(value);
	}

	/**
	 * プロパティを定義する．以下は等価である:
	 * <ul>
	 * <li><code>property(value).as(key)</code></li>
	 * <li><code>property(key, value)</code></li>
	 * </ul>
	 * @param value 値
	 * @return キーを指定するためのオブジェクト
	 * @see #property(String, Object)
	 */
	protected final AsElement property(Object value)
	{
		values.add(value);
		return asElement;
	}

	/**
	 * 内部プロパティリスト(key)を設定する．
	 */
	final void setInternalKeys(List<String> keys)
	{
		this.keys = keys;
	}

	/**
	 * 内部プロパティリスト(value)を設定する．
	 */
	final void setInternalValues(List<Object> values)
	{
		this.values = values;
	}

}
